__author__ = "Gwyn"

# ----------------------------------------------------------------------------------------------------------------------
# Name:        evaluation_functions.py
# Purpose:     defines functions used to calculate the evaluation metrics
# Author:      Gwyneth Matthews
# Created:     07/01/2021
# ----------------------------------------------------------------------------------------------------------------------

# import modules
import pandas as pd
import datetime as dt
import numpy as np
from scipy.interpolate import interp1d
from scipy.integrate import quad


# ----------------------------------------------------------------------------------------------------------------------
def extract_trajectory(fcst=None, pp_or_raw=None, trajectory=None):
    """
    Extract a single percentile trajectory  of the forecast (e.g. the median - (q50 of the post-processed forecast/6th
    member of the raw forecast).
    :param fcst: forecast data in a pandas Dataframe with dates as index vales and q1 to q99 (post-processed)
    or 'ranked_member_1' to 'ranked_member_11' (raw) as column names.
    :param pp_or_raw: 'pp' or 'raw' to define the type of forecast.
    :param trajectory: trajectory of interest - 'median', 'lowest', 'highest'
    :return: the discharge values for the quantile for each date as a pd.Dataframe
    """
    if pp_or_raw == 'raw':
        prefix = 'ranked_member_'
        if trajectory == 'median':  # 'ranked_member_6'
            quantile = 6
        elif trajectory == 'lowest':  # 'ranked_member_1'
            quantile = 1
        elif trajectory == 'highest':  # 'ranked_member_11'
            quantile = 11

    elif pp_or_raw == 'pp':
        prefix = 'q'
        if trajectory == 'median':  # q50
            quantile = 50
        elif trajectory == 'lowest':  # q1
            quantile = 1
        elif trajectory == 'highest':  # q99
            quantile = 99
    else:
        print('Incorrect input for pp_or_raw. Please use "pp" for post-processed forecasts or "raw" for raw forecasts')
        return

    return fcst[prefix + str(quantile)]


# ----------------------------------------------------------------------------------------------------------------------

def prob_exceed_threshold(fcst=None, pp_or_raw=None, thresholds=None):
    """
    Calculates the forecast probability of the discharge exceeding the MQ and MHQ thresholds at each lead-time.
    :param fcst: forecast data in a pandas Dataframe with dates as index vales and q1 to q99 (post-processed)
    or 'ranked_member_1' to 'ranked_member_11' (raw) as column names.
    :param pp_or_raw: 'pp' or 'raw' to define the type of forecast.
    :param thresholds: the MQ and MHQ thresholds for a station as a pandas Dataframe with MQ and MHQ as column names
    :return: probability of the thresholds being exceeded as a pandas Dataframe with MQ and MHQ as column names and
    dates as index values
    """
    datelist = fcst.index
    threshold_values = np.array([thresholds['MQ'].values, thresholds['MHQ'].values]).flatten()
    
    # temporary array for storing results
    exceedance_probs = np.full(
        (len(datelist), len(threshold_values)), fill_value=np.nan
    )

    # cycle through each date in the forecast period
    for date in range(len(datelist)):
        # extract the forecast values
        quantiles = fcst.loc[datelist[date]].values
        if pp_or_raw == 'raw':
            # raw probability of exceeding = (number of members above the threshold)/11
            for th in range(len(threshold_values)):
                if threshold_values[th] != np.nan:
                    exceedance_probs[date, th] = np.sum(quantiles >= threshold_values[th], axis=0) / 11.0

        elif pp_or_raw == 'pp':
            # defines linear interpolation between forecast quantiles
            f = interp1d(
                x=quantiles,
                y=np.arange(0.01, 1, 0.01),
                kind="linear",
                bounds_error=False,
                fill_value=(0, 1),
            )
            # pp probability of exceeding = 1 - probability of not exceeding
            exceedance_probs[date, ~np.isnan(threshold_values)] = 1 - f(threshold_values[~np.isnan(threshold_values)])
        else:
            print('Incorrect input for pp_or_raw. '
                  'Please use "pp" for post-processed forecasts or "raw" for raw forecasts')
            return

    return pd.DataFrame(exceedance_probs, index=datelist, columns=['MQ', 'MHQ'])


# ----------------------------------------------------------------------------------------------------------------------


def obs_prob_exceed_threshold(obs=None, thresholds=None):
    """
    Determines whether the threshold is exceeded by the observations.
    :param obs: observations in a pandas Dataframe with dates as index vales
    :param thresholds: the MQ and MHQ thresholds for a station as a pandas Dataframe with MQ and MHQ as column names
    :return: 0 if not exceeded and 1 if exceeded as a pandas Dataframe with MQ and MHQ as column names and
    dates as index values
    """
    datelist = obs.index
    threshold_values = np.array([thresholds['MQ'].values, thresholds['MHQ'].values]).flatten()

    obs_exceeding = np.zeros((len(datelist), len(threshold_values)))

    for th in range(len(threshold_values)):
        obs_exceeding[:, th] = np.array(
            obs.values.flatten() >= threshold_values[th], dtype=int
        )

    return pd.DataFrame(obs_exceeding, index=datelist, columns=['MQ', 'MHQ'])


# ----------------------------------------------------------------------------------------------------------------------


def pp_calculate_crps(cdfx=None, cdfy=None, obs=None):
    """
    Calculates the Continuous Ranked Probability Score (CRPS) for a single timestep of a post-processed forecast.
    :param cdfx: the quantiles of the forecast at the timestep of interest
    :param cdfy: np.arange(0.01, 1, 0.01)
    :param obs: value of the observation at the timstep of interest
    :return: the CRPS value at the timestep of interest
    """
    # truncate the probability distribution at 0.01 and 0.99
    close_pdf_left = cdfx[0]
    close_pdf_right = cdfx[-1]
    # define the closed pdf (no tails to infinity)
    cdfx = np.concatenate([[close_pdf_left], cdfx, [close_pdf_right]])
    cdfy = np.concatenate([[0], cdfy, [1]])

    if obs < cdfx[0]:  # forecast is completely above the obs
        # below the lower limit of ppfcst
        crps1 = cdfx[0] - obs

        f2 = interp1d(
            x=cdfx, y=(1 - cdfy) ** 2, bounds_error=True, fill_value=(1, 0)
        )  # approx. of (CDF(f) - CDF(o))^2
        crps2 = quad(
            func=f2, a=cdfx[0], b=cdfx[-1], limit=1000
        )  # integration over (CDF(f) - CDF(o))^2

        crps2 = crps2[0]

    elif obs > cdfx[-1]:  # forecast is completely below the obs
        # above the upper limit of ppfcst
        crps1 = obs - cdfx[-1]

        f2 = interp1d(
            x=cdfx, y=cdfy ** 2, bounds_error=True, fill_value=(0, 1)
        )  # approx. of (CDF(f) - CDF(o))^2
        crps2 = quad(
            func=f2, a=cdfx[0], b=cdfx[-1], limit=1000
        )  # integration over (CDF(f) - CDF(o))^2

        crps2 = crps2[0]

    elif cdfx[-1] > obs > cdfx[0]:  # obs is within the forecast spread
        # below the observation
        f1 = interp1d(
            x=cdfx, y=cdfy ** 2, bounds_error=True, fill_value=(0, 1)
        )  # approx. of (CDF(f) - CDF(o))^2
        crps1 = quad(
            func=f1, a=cdfx[0], b=obs, limit=1000
        )  # integration over (CDF(f) - CDF(o))^2
        crps1 = crps1[0]

        # above the observation
        f2 = interp1d(
            x=cdfx, y=(1 - cdfy) ** 2, bounds_error=True, fill_value=(1, 0)
        )  # approx. of (CDF(f) - CDF(o))^2
        crps2 = quad(
            func=f2, a=obs, b=cdfx[-1], limit=1000
        )  # integration over (CDF(f) - CDF(o))^2

        crps2 = crps2[0]

    else:
        print("Observation location unidentifiable.")

    return crps1 + crps2


def raw_calculate_crps(ensemble_members=None, obs=None):
    """
    Calculates the Continuous Ranked Probability Score (CRPS) for a single timestep of a raw forecast.
    :param ensemble_members: values of each of the ensemble members in ascending order for the timestep of interest.
    :param obs: value of the observation for the timesep of interest.
    :return: the CRPS value at the timestep of interest
    """

    m = len(ensemble_members)  # ensemble size
    i = np.arange(1, m + 1)  # i-th member
    member_err = ensemble_members - obs  # error of each member from the obs
    above = np.array(
        (ensemble_members >= obs), dtype=int
    )  # splits into members above and below the observation

    # calculate the crps
    a = (m * above) - i + 0.5
    b = member_err * a
    c = np.sum(b)

    crps = (2.0 / (m ** 2)) * c

    return crps


def forecast_crps(fcst=None, pp_or_raw=None, obs=None):
    """
    Calculates the CRPS for all timesteps in the forecast.
    :param fcst: forecast data in a pandas Dataframe with dates as index vales and q1 to q99 (post-processed)
    or 'ranked_member_1' to 'ranked_member_11' (raw) as column names.
    :param pp_or_raw: 'pp' or 'raw' to define the type of forecast.
    :param obs: Values of the observations at each timestep.
    :return: crps values as a pandas Dataframe with the dates as index values
    """
    datelist = fcst.index
    # array to stor results
    crps = np.full(len(datelist), fill_value=np.nan)
    # cycle through each date
    for date in range(len(datelist)):
        # if the observation is available for this date
        if not np.isnan(obs.loc[datelist[date]].values):
            if pp_or_raw == 'raw':
                crps[date] = raw_calculate_crps(
                    ensemble_members=fcst.loc[datelist[date]].values,
                    obs=obs.loc[datelist[date]].values, )

            elif pp_or_raw == 'pp':
                crps[date] = pp_calculate_crps(
                    cdfx=fcst.loc[datelist[date]].values,
                    cdfy=np.arange(0.01, 1, 0.01),
                    obs=obs.loc[datelist[date]].values)
            else:
                print(
                    'Incorrect input for pp_or_raw. '
                    'Please use "pp" for post-processed forecasts or "raw" for raw forecasts')
                return

    crps_df = pd.DataFrame(
        crps, index=datelist, columns=["crps"]
    )

    return crps_df

# ----------------------------------------------------------------------------------------------------------------------


def pte(fcst=None, pp_or_raw=None, obs=None, fcst_date=None):
    """
    Calculates the Peak-Time Error and the error in the peak magnitude for a forecast.
    :param fcst: forecast data in a pandas Dataframe with dates as index vales and q1 to q99 (post-processed)
    or 'ranked_member_1' to 'ranked_member_11' (raw) as column names.
    :param pp_or_raw: 'pp' or 'raw' to define the type of forecast.
    :param obs: values of observations at all lead-times
    :param fcst_date: date the forecast was produced.
    :return: a dictionary containing the PTE and peak magnitude error of the forecast median, the upper and the lower
    trajectories.
    """

    fcst_date = dt.datetime.strptime(fcst_date, "%Y%m%d%H")
    max_values = np.sort(fcst.max().values)
    max_times = np.sort(fcst.idxmax().values)

    pte_dict = {}
    quant = [50, 0, 100]

    if pp_or_raw == "raw":
        q_idx = [5, 0, 10]
    elif pp_or_raw == "pp":
        q_idx = [49, 0, 98]

    # find the maximum value (p) and the timing of the maximum value (t) for each of the defined quantiles
    # median trajectory
    pte_dict["p" + str(quant[0])] = max_values[q_idx[0]]
    pte_dict["t" + str(quant[0])] = (
        dt.datetime.strptime(str(max_times[q_idx[0]]), "%Y%m%d%H") - fcst_date
    ).days
    # lower trajectory
    pte_dict["p" + str(quant[1])] = max_values[q_idx[1]]
    pte_dict["t" + str(quant[1])] = (
        dt.datetime.strptime(str(max_times[q_idx[1]]), "%Y%m%d%H") - fcst_date
    ).days
    # upper trajectory
    pte_dict["p" + str(quant[2])] = max_values[q_idx[2]]
    pte_dict["t" + str(quant[2])] = (
        dt.datetime.strptime(str(max_times[q_idx[2]]), "%Y%m%d%H") - fcst_date
    ).days

    # find the maximum value and the timing of the maximum value for the observations
    pte_dict["p_obs"] = obs.max().values[0]
    pte_dict["t_obs"] = (
        dt.datetime.strptime(str(obs.idxmax().values[0]), "%Y%m%d%H") - fcst_date
    ).days

    # find the difference between the forecast maximum values and the observed maximum values
    # median trajectory
    pte_dict["D_peak_" + str(quant[0])] = pte_dict["p" + str(quant[0])] - obs.max().values[0]
    pte_dict["D_time_" + str(quant[0])] = (
        pte_dict["t" + str(quant[0])] - (dt.datetime.strptime(str(obs.idxmax().values[0]), "%Y%m%d%H") - fcst_date).days
    )
    # lower trajectory
    pte_dict["D_peak_" + str(quant[1])] = pte_dict["p" + str(quant[1])] - obs.max().values[0]

    pte_dict["D_time_" + str(quant[1])] = (
        pte_dict["t" + str(quant[1])] - (dt.datetime.strptime(str(obs.idxmax().values[0]), "%Y%m%d%H") - fcst_date).days
    )
    # upper trajectory
    pte_dict["D_peak_" + str(quant[2])] = (
        pte_dict["p" + str(quant[2])] - obs.max().values[0]
    )
    pte_dict["D_time_" + str(quant[2])] = (
        pte_dict["t" + str(quant[2])]
        - (
            dt.datetime.strptime(str(obs.idxmax().values[0]), "%Y%m%d%H") - fcst_date
        ).days
    )

    return pte_dict
