__author__ = 'Gwyneth Matthews'

# ----------------------------------------------------------------------------------------------------------------------
# Name:        evaluation_of_forecasts.py
# Purpose:     calculate all metrics for all forecasts
# Author:      Gwyneth Matthews
# Created:     07/01/2021
# ----------------------------------------------------------------------------------------------------------------------

# import modules
import numpy as np
import pandas as pd
import os
import evaluation_functions as ef
from scipy.stats import pearsonr


def open_fcst(path=None):
    """
    Opens the csv file at the given path location and returns the forecast contained.
    :param path: path to .csv file containing the forecast of interest.
    :return: forecast as a pandas Dataframe

    """
    return pd.read_csv(path, index_col=0)


def open_thresholds(station=None):
    """
    Extracts the MQ, MHQ, and 10% threshold values from the station_data.csv file for a single station.
    :param station: stations of interest
    :return: thresholds as a pandas Dataframe
    """
    threshold_data = pd.read_csv("../Data/station_data.csv", usecols=['stationID', 'MQ', 'MHQ', '10%'])
    return threshold_data[threshold_data['stationID'] == station]


def rank_raw_members(fcst=None):
    """
    Rank the raw forecast ensemble members at each lead-time.
    :param fcst: raw forecast as a pandas Dataframe with the dates as index values
    :return: forecast as a pandas Dataframe with the members ranked at each timestep
    """
    datelist = fcst.index
    ranked_fcst = np.full((len(datelist), 11), fill_value=np.nan)
    for date in range(len(datelist)):
        ranked_fcst[date] = np.sort(fcst.loc[datelist[date]].values)

    ranked_fcst = pd.DataFrame(
        ranked_fcst,
        index=datelist,
        columns=["ranked_member_%i" % i for i in range(1, 12)],
    )

    return ranked_fcst


def collate_forecasts(pp_or_raw=None):
    """
    Collates all station forecasts for each forecast date into a single file. This is done for efficiency.
    :param pp_or_raw: 'pp' or 'raw' to define forecast set of interest
    :return: saved .npy file for each forecast date containing a dictionary with an item per station.
    """
    # define if raw or post-processed forecast
    if pp_or_raw == 'raw':
        forecast_str = 'raw_reforecasts'
    elif pp_or_raw == 'pp':
        forecast_str = 'pp_reforecasts'
    else:
        print('Incorrect input for pp_or_raw. Please use "pp" for post-processed forecasts or "raw" for raw forecasts')

    # path to forecast data
    fcst_path = "../Data/" + forecast_str + "/"

    # open station IDs and forecast dates
    stations = np.array(pd.read_csv("../Data/station_data.csv")['stationID'].values, dtype=int)
    fcst_dates = np.loadtxt(
        "../Data/forecast_dates.txt")
    fcst_dates = fcst_dates.astype(int).astype(str)

    # cycle through reforecast dates
    for fcst_date in fcst_dates:
        # create a dictionary to contain the forecasts
        fcst_dict = {}

        # cycle through the stations
        for stn in stations:
            # path to forecast
            fcst_path = fcst_path + fcst_date + '/' + forecast_str[:-1] + '_' + fcst_date + '_' + str(stn) + '.csv'
            # open forecast
            fcst = open_fcst(fcst_path)
            # store forecast
            fcst_dict[str(int(stn))] = fcst
        # save forecast
        np.save(fcst_path + forecast_str[:-1] + '_' + fcst_date, fcst_dict, allow_pickle=True)


def evaluate_single_forecast(
    fcst=None,
    results_path=None,
    pp_or_raw=None,
    station=None,
    date=None,
):
    """
        Calculates the crps, pte, probability of exceedance, and extracts the forecast median for the station forecast
        of interest.
        :param fcst: forecast data in a pandas Dataframe with dates as index vales and q1 to q99 (post-processed)
        or 'ranked_member_1' to 'ranked_member_11' (raw) as column names.
        :param results_path: path to where results should be saved.
        :param pp_or_raw: 'pp' or 'raw' to define the type of forecast.
        :param station: station ID for station of interest.
        :param date: date forecast was produced as a string.
        :return:
        """
    if pp_or_raw == 'raw':
        fcst = rank_raw_members(fcst=fcst)

    thresholds = open_thresholds(station=station)

    datelist = fcst.index
    # open observations file
    obs_path = "../Data/observations/" + date + "/obs_" + date + "_" + str(station) + ".npy"
    obs = np.load(obs_path)
    obs = pd.DataFrame(obs, index=datelist)

    # extract forecast median
    median_traj = ef.extract_trajectory(fcst=fcst, pp_or_raw=pp_or_raw, trajectory='median')
    np.save(
        results_path + "median_trajectory_" + date + "_" + str(station), median_traj
    )

    # probability of threshold exceedance
    prob_exceed_df = ef.prob_exceed_threshold(
        fcst=fcst, pp_or_raw=pp_or_raw, thresholds=thresholds
    )

    np.save(
        results_path + "probability_of_exceedance_" + date + "_" + str(station), prob_exceed_df,
    )

    # CRPS
    crps = ef.forecast_crps(fcst=fcst, pp_or_raw=pp_or_raw, obs=obs)
    np.save(
        results_path + 'crps_' + date + '_' + str(station), crps
    )

    # Peak-Time Error
    if np.nanmax(obs.values[:]) >= thresholds["10%"].values:
        pte_dict = ef.pte(
            raw_or_pp=pp_or_raw,
            fcst=fcst,
            obs=obs,
            fcst_date=date
        )
        np.save(
            results_path + "PTE_" + date + "_" + str(station), pte_dict
        )


def evaluate_all_forecasts(pp_or_raw=None):
    # get stations and forecasts numbers
    stations = np.array(pd.read_csv("../Data/station_data.csv")['stationID'].values, dtype=int)
    fcst_dates = np.loadtxt("../Data/forecast_dates.txt")
    fcst_dates = fcst_dates.astype(int).astype(str)

    if pp_or_raw == 'raw':
        forecast_str = 'raw_reforecasts'
    elif pp_or_raw == 'pp':
        forecast_str = 'pp_reforecasts'
    else:
        print('Incorrect input for pp_or_raw. Please use "pp" for post-processed forecasts or "raw" for raw forecasts')

    # iterate over all forecasts
    for fcst_date in fcst_dates:

        # set up results files for this forecast
        if not os.path.isdir("../Data/" + forecast_str[:-12] + "_results/" + fcst_date):
            os.mkdir("../Data/" + forecast_str[:-12] + "_results/" + fcst_date)

        # open forecast data
        path_to_forecast = "../Data/" + forecast_str + "/" + forecast_str[:-1] + "_" + fcst_date + ".npy"
        full_data = np.load(path_to_forecast, allow_pickle=True).flat[0]

        # cycle through each station
        for stn in stations:
            # set up results files for this station
            results_path = "../Data/" + forecast_str[:-12] + "_results/" + fcst_date + "/" + str(stn)
            if not os.path.isdir(results_path):
                os.mkdir(results_path)

            # create pd.DataFrame of forecast for the station
            stn_fcst = full_data[str(int(stn))]

            # define paths for station
            results_path = results_path + "/"

            # evaluate station forecast
            evaluate_single_forecast(
                fcst=stn_fcst,
                results_path=results_path,
                pp_or_raw=pp_or_raw,
                station=stn,
                date=fcst_date,
            )


def collate_results(pp_or_raw=None):
    """
    Collects all results for all stations and forecasts into single files.
    :param pp_or_raw: 'pp' or 'raw' to define the type of forecast.
    :return:
    """
    # set base path for all forecast results
    if pp_or_raw == 'raw':
        forecast_str = 'raw_results/'
    elif pp_or_raw == 'pp':
        forecast_str = 'pp_results/'
    else:
        print('Incorrect input for pp_or_raw. Please use "pp" for post-processed forecasts or "raw" for raw forecasts')

    results_path = "../Data/" + forecast_str
    # get stations and forecasts numbers
    stations = np.array(pd.read_csv("../Data/station_data.csv")['stationID'].values, dtype=int)
    fcst_dates = np.loadtxt("../Data/forecast_dates.txt")
    fcst_dates = fcst_dates.astype(int).astype(str)

    probs = np.arange(0, 1.1, 0.05)

    # arrays to store results
    medians = np.full((len(stations), 15, len(fcst_dates)), fill_value=np.nan)
    observations = np.full((len(stations), 15, len(fcst_dates)), fill_value=np.nan)
    crps_values = np.full((len(stations), 15, len(fcst_dates)), fill_value=np.nan)
    d_time_values = np.full((len(stations), len(fcst_dates), 3), fill_value=np.nan)
    obs_time_values = np.full((len(stations), len(fcst_dates), 3), fill_value=np.nan)
    d_peak_values = np.full((len(stations), len(fcst_dates), 3), fill_value=np.nan)
    hits = np.full((len(probs), 2, 3), fill_value=0)
    misses = np.full((len(probs), 2, 3), fill_value=0)
    false_alarms = np.full((len(probs), 2, 3), fill_value=0)
    correct_negs = np.full((len(probs), 2, 3), fill_value=0)
    forecast_probs = np.full((len(probs), 2, 3), fill_value=0)
    observed_events = np.full((len(probs), 2, 3), fill_value=0)

    # cycle through forecasts
    for fcst_date in range(len(fcst_dates)):
        # cycle through stations
        for stn in range(len(stations)):

            date = fcst_dates[fcst_date]
            station = stations[stn]
            thresholds = open_thresholds(station)
            threshold_names = ['MQ', 'MHQ']

            # collate observations
            obs_path = "../Data/observations/" + date + "/obs_" + date + "_" + str(station) + ".npy"
            obs = np.load(obs_path)
            obs_avail = ~np.isnan(obs.flatten())
            obs = obs.flatten()[obs_avail]
            observations[stn, obs_avail, fcst_date] = obs.flatten()

            # collate forecast medians
            med_path = results_path + '%s/%i/median_trajectory_%s_%i.npy' % (date, station, date, station)
            median = np.load(med_path)
            medians[stn, obs_avail, fcst_date] = median.flatten()[obs_avail]

            # collate crps values
            crps_path = results_path + '%s/%i/crps_%s_%i.npy' % (date, station, date, station)
            crps = np.load(crps_path)
            crps_values[stn, obs_avail, fcst_date] = crps.flatten()[obs_avail]

            # collate Peak-Time error data
            pte_dict_path = (results_path + "%s/%i/PTE_%s_%i.npy" % (date, station, date, station))
            if os.path.isfile(pte_dict_path):
                pte_dict = np.load(pte_dict_path, allow_pickle=True).flat[0]

                d_time_values[stn, fcst_date, 0] = pte_dict['D_time_50']
                d_peak_values[stn, fcst_date, 0] = pte_dict['D_peak_50']
                d_time_values[stn, fcst_date, 1] = pte_dict['D_time_0']
                d_peak_values[stn, fcst_date, 1] = pte_dict['D_peak_0']
                d_time_values[stn, fcst_date, 2] = pte_dict['D_time_100']
                d_peak_values[stn, fcst_date, 2] = pte_dict['D_peak_100']
                obs_time_values[stn, fcst_date] = pte_dict['t_obs']

            # collate
            poe_path = results_path + '%s/%i/probability_of_exceedance_%s_%i.npy' % (date, station, date, station)
            poe_full = np.load(poe_path)
            poe = poe_full[obs_avail]
            for th in [0, 1]:
                thresh = thresholds[threshold_names[th]].values
                # events; where the observations exceed the threshold
                event_obs = np.array(obs.flatten() > thresh, dtype=int)

                event_obs_bool = np.array(event_obs, dtype=bool)

                for p in range(len(probs)):

                    values_hits = np.zeros(15)[obs_avail]
                    values_misses = np.zeros(15)[obs_avail]
                    values_fa = np.zeros(15)[obs_avail]
                    values_cn = np.zeros(15)[obs_avail]

                    if np.sum(event_obs) > 0:
                        values_hits[event_obs_bool] = np.array(poe[event_obs_bool, th] >= probs[p], dtype=int)
                        values_misses[event_obs_bool] = np.array(poe[event_obs_bool, th] < probs[p], dtype=int)
                        values_fa[~event_obs_bool] = np.array(poe[~event_obs_bool, th] >= probs[p], dtype=int)
                        values_cn[~event_obs_bool] = np.array(poe[~event_obs_bool, th] < probs[p], dtype=int)
                    else:
                        values_fa[~event_obs_bool] = np.array(poe[~event_obs_bool, th] >= probs[p], dtype=int)
                        values_cn[~event_obs_bool] = np.array(poe[~event_obs_bool, th] < probs[p], dtype=int)

                    for i in [0, 1, 2]:
                        hits[p, th, i] += np.nansum(
                            values_hits[i * 5:(i + 1) * 5][~np.isnan(values_hits[i * 5:(i + 1) * 5])])
                        misses[p, th, i] += np.nansum(
                            values_misses[i * 5:(i + 1) * 5][~np.isnan(values_misses[i * 5:(i + 1) * 5])])
                        false_alarms[p, th, i] += np.nansum(
                            values_fa[i * 5:(i + 1) * 5][~np.isnan(values_fa[i * 5:(i + 1) * 5])])
                        correct_negs[p, th, i] += np.nansum(
                            values_cn[i * 5:(i + 1) * 5][~np.isnan(values_cn[i * 5:(i + 1) * 5])])

                k = 0
                for i in np.arange(15):
                    poe_th = poe_full[i, th]
                    
                    if obs_avail[i]:
                        if i < 5:
                            if poe_th == 0:
                                forecast_probs[0, th, 0] += 1
                                observed_events[0, th, 0] += int(event_obs[k])
                            else:
                                in_bin = np.searchsorted(probs, poe_th, side='right')
                                forecast_probs[in_bin, th, 0] += 1
                                observed_events[in_bin, th, 0] += int(event_obs[k])

                        elif i in [5, 6, 7, 8, 9]:
                            if poe_th == 0:
                                forecast_probs[0, th, 1] += 1
                                observed_events[0, th, 1] += int(event_obs[k])
                            else:
                                in_bin = np.searchsorted(probs, poe_th, side='right')
                                forecast_probs[in_bin, th, 1] += 1
                                observed_events[in_bin, th, 1] += int(event_obs[k])

                        elif i > 9:
                            if poe_th == 0:
                                forecast_probs[0, th, 2] += 1
                                observed_events[0, th, 2] += int(event_obs[k])
                            else:
                                in_bin = np.searchsorted(probs, poe_th, side='right')
                                forecast_probs[in_bin, th, 2] += 1
                                observed_events[in_bin, th, 2] += int(event_obs[k])
                        k += 1

    # save all results
    np.save(results_path + 'medians_all_fcasts_all_stns', medians)
    np.save(results_path + 'obs_all_fcasts_all_stns', observations)
    np.save(results_path + 'crps_all_fcasts_all_stns', crps_values)
    np.save(results_path + 'PTE_peaks_all_fcasts_all_stns', d_peak_values)
    np.save(results_path + 'PTE_times_all_fcasts_all_stns', d_time_values)
    np.save(results_path + 'PTE_obs_times_all_fcasts_all_stns', obs_time_values)

    np.save(results_path + 'hits_all_fcasts_all_stns', hits)
    np.save(results_path + 'misses_all_fcasts_all_stns', misses)
    np.save(results_path + 'false_alarms_all_fcasts_all_stns', false_alarms)
    np.save(results_path + 'correct_negatives_all_fcasts_all_stns', correct_negs)
    np.save(results_path + 'forecast_probs', forecast_probs)
    np.save(results_path + 'observed_events', observed_events)


def calculate_kge(pp_or_raw=None):
    """
    Calculates the modified KGE for each lead-time and each station.
    :param pp_or_raw: 'pp' or 'raw' to define the type of forecast.
    :return:
    """
    if pp_or_raw == 'raw':
        forecast_str = 'raw_results/'
    elif pp_or_raw == 'pp':
        forecast_str = 'pp_results/'
    else:
        print('Incorrect input for pp_or_raw. Please use "pp" for post-processed forecasts or "raw" for raw forecasts')

    results_path = "../Data/" + forecast_str
    stations = np.array(pd.read_csv("../Data/station_data.csv")['stationID'].values, dtype=int)
    # array to store results
    kge_values = np.full((len(stations), 15), fill_value=np.nan)
    r_values = np.full((len(stations), 15), fill_value=np.nan)
    beta_values = np.full((len(stations), 15), fill_value=np.nan)
    gamma_values = np.full((len(stations), 15), fill_value=np.nan)
    medians = np.load(results_path + 'medians_all_fcasts_all_stns.npy')
    observations = np.load(results_path + 'obs_all_fcasts_all_stns.npy')

    for stn in range(len(stations)):
        for lt in range(15):
            obs_timeseries = observations[stn, lt, :]
            sim_timeseries = medians[stn, lt, :]
            sim_timeseries = sim_timeseries[~np.isnan(obs_timeseries)]
            obs_timeseries = obs_timeseries[~np.isnan(obs_timeseries)]

            if len(obs_timeseries) < 3:
                continue
            r = pearsonr(sim_timeseries, obs_timeseries)
            # means
            mu_fcst = np.mean(sim_timeseries)
            mu_obs = np.mean(obs_timeseries)

            # standard deviations
            sigma_fcst = np.std(sim_timeseries)
            sigma_obs = np.std(obs_timeseries)

            # beta
            beta = mu_fcst / mu_obs

            # gamma
            gamma = (sigma_fcst / mu_fcst) / (sigma_obs / mu_obs)

            ed2 = (r[0] - 1) ** 2 + (beta - 1) ** 2 + (gamma - 1) ** 2
            kge = 1 - np.sqrt(ed2)
            kge_values[stn, lt] = kge
            r_values[stn, lt] = r[0]
            beta_values[stn, lt] = beta
            gamma_values[stn, lt] = gamma

    # save results
    np.save(results_path + 'KGE_all_fcasts_all_stns', kge_values)
    np.save(results_path + 'r_all_fcasts_all_stns', r_values)
    np.save(results_path + 'beta_all_fcasts_all_stns', beta_values)
    np.save(results_path + 'gamma_all_fcasts_all_stns', gamma_values)


def calculate_crps(pp_or_raw=None):
    """
    Calculates the average CRPS for each stations and each lead-time.
    :param pp_or_raw: 'pp' or 'raw' to define the type of forecast.
    :return:
    """
    if pp_or_raw == 'raw':
        forecast_str = 'raw_results/'
    elif pp_or_raw == 'pp':
        forecast_str = 'pp_results/'
    else:
        print('Incorrect input for pp_or_raw. Please use "pp" for post-processed forecasts or "raw" for raw forecasts')

    results_path = "../Data/" + forecast_str
    crps_values = np.load(results_path + 'crps_all_fcasts_all_stns.npy')
    avg_crps_values = np.nanmean(crps_values, axis=2).T

    # save results
    np.save(results_path + 'avg_crps_all_stns', avg_crps_values)


def calculate_roc(pp_or_raw=None):
    """
    Calculates the hit rate and the false alarm ratio for use in the ROC diagram.
    :param pp_or_raw: 'pp' or 'raw' to define the type of forecast.
    :return:
    """
    if pp_or_raw == 'raw':
        forecast_str = 'raw_results/'
    elif pp_or_raw == 'pp':
        forecast_str = 'pp_results/'
    else:
        print('Incorrect input for pp_or_raw. Please use "pp" for post-processed forecasts or "raw" for raw forecasts')

    results_path = "../Data/" + forecast_str
    hit = np.load(results_path + 'hits_all_fcasts_all_stns.npy')
    miss = np.load(results_path + 'misses_all_fcasts_all_stns.npy')
    fa = np.load(results_path + 'false_alarms_all_fcasts_all_stns.npy')
    cn = np.load(results_path + 'correct_negatives_all_fcasts_all_stns.npy')

    pod = hit / (hit + miss)
    far = fa / (fa + cn)

    # save results
    np.save(results_path + 'probability_of_detection', pod)
    np.save(results_path + 'false_alarm_rate', far)


def full_evaluation(pp_or_raw=None):
    """
    Runs the full evaluation and collating of results for a set of forecasts.
    :param pp_or_raw: 'pp' or 'raw' to define the type of forecast.
    :return:
    """
    evaluate_all_forecasts(pp_or_raw=pp_or_raw)
    collate_results(pp_or_raw=pp_or_raw)
    calculate_kge(pp_or_raw=pp_or_raw)
    calculate_crps(pp_or_raw=pp_or_raw)
    calculate_roc(pp_or_raw=pp_or_raw)
